//! Review requests interface
use serde::{Deserialize, Serialize};

use crate::pulls::Pull;
use crate::teams::Team;
use crate::users::User;
use crate::{Future, Github};

/// A structure for interfacing with review requests
pub struct ReviewRequests {
    github: Github,
    owner: String,
    repo: String,
    number: u64,
}

impl ReviewRequests {
    #[doc(hidden)]
    pub fn new<O, R>(github: Github, owner: O, repo: R, number: u64) -> Self
    where
        O: Into<String>,
        R: Into<String>,
    {
        ReviewRequests {
            github,
            owner: owner.into(),
            repo: repo.into(),
            number,
        }
    }

    /// list requested reviews
    pub fn get(&self) -> Future<ReviewRequest> {
        self.github.get::<ReviewRequest>(&self.path())
    }

    /// Add new requested reviews
    pub fn create(&self, review_request: &ReviewRequestOptions) -> Future<Pull> {
        self.github.post(&self.path(), json!(review_request))
    }

    /// Delete a review request
    pub fn delete(&self, review_request: &ReviewRequestOptions) -> Future<()> {
        self.github
            .delete_message(&self.path(), json!(review_request))
    }

    fn path(&self) -> String {
        format!(
            "/repos/{}/{}/pulls/{}/requested_reviewers",
            self.owner, self.repo, self.number
        )
    }
}

// representations (todo: replace with derive_builder)

#[derive(Default, Serialize)]
pub struct ReviewRequestOptions {
    /// An array of user `logins` that will be requested.
    /// Note, each login must be a collaborator.
    pub reviewers: Vec<String>,
    /// An array of team `slugs` that will be requested.
    pub team_reviewers: Vec<String>,
}

#[derive(Debug, Deserialize)]
pub struct ReviewRequest {
    pub users: Vec<User>,
    pub teams: Vec<Team>,
}
